import { Mutation, isObject, isDefined, hashIterator, unwrap, unwrapKey } from '@collectable/core';
import { RedBlackTree } from '@collectable/red-black-tree';
import { HashMap } from '@collectable/map';
import { setItem } from './values';
import { isEqual, get, set, has, update } from '../functions';
import { iteratePairs } from './iterate';
export class SortedMapStructure {
    /** @internal */
    constructor(mctx, _indexed, _sorted, _compare, _select) {
        this._indexed = _indexed;
        this._sorted = _sorted;
        this._compare = _compare;
        this._select = _select;
        this['@@mctx'] = mctx;
    }
    /** @internal */
    get '@@size'() { return this._indexed['@@size']; }
    /** @internal */
    get '@@is-collection'() { return true; }
    /** @internal */
    '@@clone'(mctx) {
        var sctx = Mutation.asSubordinateContext(mctx);
        return new SortedMapStructure(mctx, Mutation.clone(this._indexed, sctx), Mutation.clone(this._sorted, sctx), this._compare, this._select);
    }
    /** @internal */
    '@@equals'(other) {
        return isEqual(this, other);
    }
    /** @internal */
    '@@hash'() {
        return hashIterator(iteratePairs(this));
    }
    /** @internal */
    '@@unwrap'() {
        return unwrap(this);
    }
    /** @internal */
    '@@unwrapInto'(target) {
        var it = RedBlackTree.iterateFromFirst(this._sorted);
        var current;
        while (!(current = it.next()).done) {
            var entry = current.value;
            target[unwrapKey(entry.key.key)] = unwrap(entry.value);
        }
        return target;
    }
    /** @internal */
    '@@createUnwrapTarget'() {
        return {};
    }
    /** @internal */
    '@@get'(key) {
        return get(key, this);
    }
    /** @internal */
    '@@has'(key) {
        return has(key, this);
    }
    /** @internal */
    '@@set'(key, value) {
        return set(key, value, this);
    }
    /** @internal */
    '@@update'(updater, key) {
        return update(updater, key, this);
    }
    /** @internal */
    '@@verifyKey'(key) {
        return true;
    }
    [Symbol.iterator]() {
        return iteratePairs(this);
    }
}
export function isSortedMap(arg) {
    return isObject(arg) && arg instanceof SortedMapStructure;
}
export function cloneSortedMap(map, clear, mutability) {
    var mctx = Mutation.selectContext(mutability);
    var sctx = Mutation.asSubordinateContext(mctx);
    var keys;
    var sorted;
    if (clear) {
        keys = HashMap.empty(sctx);
        sorted = RedBlackTree.empty(map._compare, sctx);
    }
    else {
        keys = Mutation.clone(map._indexed, sctx);
        sorted = Mutation.clone(map._sorted, sctx);
    }
    return new SortedMapStructure(mctx, keys, sorted, map._compare, map._select);
}
export function createMap(values, compare, select, mutability) {
    var map = Mutation.modify(emptySortedMap(compare, select, mutability));
    var indexed = map._indexed;
    var sorted = map._sorted;
    if (Array.isArray(values)) {
        for (var i = 0; i < values.length; i++) {
            var value = values[i];
            setItem(value[0], value[1], indexed, sorted, select);
        }
    }
    else {
        var it = values[Symbol.iterator]();
        var current;
        while (!(current = it.next()).done) {
            value = current.value;
            setItem(value[0], value[1], indexed, sorted, select);
        }
    }
    return Mutation.commit(map);
}
export function extractTree(set) {
    return set._sorted;
}
export function extractMap(set) {
    return set._indexed;
}
const DEFAULT_COMPARATOR = (a, b) => a.index - b.index;
const COMPARATOR_CACHE = new WeakMap();
function createComparatorFn(compare) {
    var fn = COMPARATOR_CACHE.get(compare);
    return isDefined(fn) ? fn : (fn = function (a, b) {
        var n = compare(a, b);
        return n === 0 ? DEFAULT_COMPARATOR(a, b) : n;
    }, COMPARATOR_CACHE.set(compare, fn), fn);
}
export function emptySortedMap(compare, select, mutability) {
    var comparator = isDefined(compare) ? createComparatorFn(compare) : DEFAULT_COMPARATOR;
    var mctx = Mutation.selectContext(mutability);
    if (Mutation.isImmutableContext(mctx) && comparator === DEFAULT_COMPARATOR) {
        return isDefined(EMPTY) ? EMPTY : (EMPTY = new SortedMapStructure(mctx, HashMap.empty(), RedBlackTree.empty(comparator), comparator, void 0));
    }
    var sctx = Mutation.asSubordinateContext(mctx);
    var map = HashMap.empty(sctx);
    var tree = RedBlackTree.empty(comparator, sctx);
    return new SortedMapStructure(mctx, map, tree, comparator, select);
}
var EMPTY;

//# sourceMappingURL=SortedMap.js.map
